# Copyright 2015 Alex Brandt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import typing  # flake8: noqa (use mypy typing)
import unittest
import unittest.mock

import torment

logger = logging.getLogger(__name__)

PATCH = lambda: 'not patched'


class MetaContextGenerateCasesUnitTest(unittest.TestCase):
    def setUp(self) -> None:
        _ = unittest.mock.patch('torment.fixtures.of')
        self.mocked_torment_fixtures_of = _.start()
        self.addCleanup(_.stop)

        self.directory = [
            '__class__',
            '__delattr__',
            '__dict__',
            '__dir__',
            '__doc__',
            '__eq__',
            '__format__',
            '__ge__',
            '__getattribute__',
            '__gt__',
            '__hash__',
            '__init__',
            '__le__',
            '__lt__',
            '__module__',
            '__ne__',
            '__new__',
            '__reduce__',
            '__reduce_ex__',
            '__repr__',
            '__setattr__',
            '__sizeof__',
            '__str__',
            '__subclasshook__',
            '__weakref__',
            'mocks',
            'mocks_mask',
        ]

    def test_null_fixture_classes(self) -> None:
        '''torment.MetaContext: fixture_classes == ø'''

        with self.assertWarns(UserWarning):
            class null_fixture_classes(object, metaclass = torment.MetaContext):
                pass

        self.assertCountEqual(dir(null_fixture_classes), self.directory)

    def test_zero_fixture_classes(self) -> None:
        '''torment.MetaContext: len(fixture_classes) == 0'''

        self.mocked_torment_fixtures_of.return_value = []

        class zero_fixture_classes(object, metaclass = torment.MetaContext):
            fixture_classes = ()

        self.directory.append('fixture_classes')

        self.assertCountEqual(dir(zero_fixture_classes), self.directory)

    def test_one_fixture_classes(self) -> None:
        '''torment.MetaContext: len(fixture_classes) > 0'''

        class dummy_fixture_a(object):
            @property
            def name(self) -> str:
                return 'test_method'

            @property
            def description(self) -> str:
                return 'test_method'

        self.mocked_torment_fixtures_of.return_value = [ dummy_fixture_a(), ]

        class many_fixture_classes(object, metaclass = torment.MetaContext):
            fixture_classes = ( dummy_fixture_a, )

        self.directory.extend([
            'fixture_classes',
            'test_method',
        ])

        self.assertCountEqual(dir(many_fixture_classes), self.directory)


class TestContextPropertyUnitTest(unittest.TestCase):
    def setUp(self) -> None:
        self.c = torment.TestContext()

    def test_testcontext_module(self) -> None:
        '''torment.TestContext().module == 'torment.fixtures' '''

        self.c.__module__ = 'test_torment.test_unit.test_fixtures'

        self.assertEqual(self.c.module, 'torment.fixtures')


class TestContextPatchUnitTest(unittest.TestCase):
    @unittest.skip('TODO test patch')
    def test_testcontext_patch(self) -> None:
        '''torment.TestContext().patch('PATCH')'''

        logger.debug('self.__module__: %s', self.__module__)

        mock_c = unittest.mock.MagicMock(wraps = torment.TestContext(), module = self.__module__)
        logger.debug('mock_c.module: %s', mock_c.module)

        mock_c.patch('PATCH')

        self.assertTrue(hasattr(mock_c, 'mocked_PATCH'))
        self.assertIsInstance(mock_c.mocked_PATCH, unittest.mock.MagicMock)



# Copyright 2015 Alex Brandt <alex.brandt@rackspace.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import importlib
import itertools
import logging
import os
import typing  # flake8: noqa (use mypy typing)

from typing import Any
from typing import Dict
from typing import Iterable
from typing import List

from torment import decorators

logger = logging.getLogger(__name__)


@decorators.log
def extend(base: Dict[Any, Any], extension: Dict[Any, Any]) -> Dict[Any, Any]:
    '''Extend base by updating with the extension.
 
    Arguments
    ---------

    :``base``: dictionary to have keys updated or added
    :``extension``: dictionary to update base with

    Return Value(s)
    ---------------

    Resulting dictionary from updating base with extension.

    '''

    _ = copy.deepcopy(base)
    _.update(extension)

    return _


@decorators.log
def import_directory(module_basename: str, directory: str, sort_key = None) -> None:
    '''Load all python modules in directory and directory's children.

    Parameters
    ----------

    :``module_basename``: module name prefix for loaded modules
    :``directory``:       directory to load python modules from
    :``sort_key``:        function to sort module names with before loading

    '''

    logger.info('loading submodules of %s', module_basename)
    logger.info('loading modules from %s', directory)

    filenames = itertools.chain(*[ [ os.path.join(_[0], filename) for filename in _[2] ] for _ in os.walk(directory) if len(_[2]) ])
    modulenames = _filenames_to_modulenames(filenames, module_basename, directory)

    for modulename in sorted(modulenames, key = sort_key):
        try:
            importlib.import_module(modulename)
        except ImportError:
            logger.warning('failed loading %s', modulename)
            logger.exception('module loading failure')
        else:
            logger.info('successfully loaded %s', modulename)


@decorators.log
def _filenames_to_modulenames(filenames: Iterable[str], modulename_prefix: str, filename_prefix: str = '') -> Iterable[str]:
    '''Convert given filenames to module names.

    Any filename that does not have a corresponding module name will be dropped
    from the result (i.e. __init__.py).

    Parameters
    ----------

    :``filename_prefix``:   a prefix to drop from all filenames (typically a
                            common directory); defaults to ''
    :``filenames``:         the filenames to transform into module names
    :``modulename_prefix``: a prefix to add to all module names

    Return Value(s)
    ---------------

    A list of modulenames corresponding to all filenames (for legal module names).

    '''

    modulenames = []  # type: Iterable[str]

    for filename in filenames:
        if not filename.endswith('.py'):
            continue

        name = filename

        name = name.replace(filename_prefix, '')
        name = name.replace('__init__.py', '')
        name = name.replace('.py', '')
        name = name.replace('/', '.')

        name = name.strip('.')

        if not len(name):
            continue

        if not modulename_prefix.endswith('.'):
            modulename_prefix += '.'

        name = modulename_prefix + name

        known_symbols = set()
        name = '.'.join([ _ for _ in name.split('.') if _ not in known_symbols and not known_symbols.add(_) ])

        if len(name):
            modulenames.append(name)

    return modulenames



# Copyright 2015 Alex Brandt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing  # flake8: noqa (use mypy typing)

from torment import fixtures
from torment import MetaContext
from torment import TestContext

from torment import decorators


class LogDecoratorFixture(fixtures.Fixture):
    @property
    def description(self) -> str:
        _description = super().description + '.log({0.function.__name__})'

        if self.parameters.get('prefix') is not None:
            _description = _description[:20] + '({0.parameters[prefix]})' + _description[20:]

        return _description.format(self, self.context.module)

    def run(self) -> None:
        pass  # TODO capture logs

    def check(self) -> None:
        pass  # TODO capture logs


class MockDecoratorFixture(fixtures.Fixture):
    @property
    def description(self) -> str:
        return super().description + '.mock({0.mock.name})'

    def run(self) -> None:
        pass  # TODO mock symbol

    def check(self) -> None:
        pass  # TODO mock symbol


class DecoratorUnitTest(TestContext, metaclass = MetaContext):
    fixture_classes = (
        LogDecoratorFixture,
        MockDecoratorFixture,
    )



# Copyright 2015 Alex Brandt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import inspect
import logging
import os
import sys
import typing  # flake8: noqa (use mypy typing)
import uuid

from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterable
from typing import Tuple
from typing import Union

from torment import decorators

logger = logging.getLogger(__name__)


class Fixture(object):
    '''Collection of data and actions for a particular test case.

    Intended as a base class for custom fixtures.  Fixture provides an API
    that simplifies writing scalable test cases.

    Creating Fixture objects is broken into two parts.  This keeps the logic for
    a class of test cases separate from the data for particular cases while
    allowing re-use of the data provided by a fixture.

    The first part of Fixture object creation is crafting a proper subclass that
    implements the necessary actions:

    :``__init__``:   pre-data population initialization
    :``initialize``: post-data population initialization
    :``setup``:      pre-run setup
    :``run``:        REQUIRED—run code under test
    :``check``:      verify results of run

    .. note::
        ``initialize`` is run during ``__init__`` and setup is run after;
        otherwise, they serve the same function.  The split allows different
        actions to occur in different areas of the class heirarchy and generally
        isn't necessary.

    By default all actions are noops and simply do nothing but run is required.
    These actions allow complex class hierarchies to provide nuanced testing
    behavior.  For example, Fixture provides the absolute bare minimum to test
    any Fixture and no more.  By adding a set of subclasses, common
    initialization and checks can be performed at one layer while specific run
    decisions and checks can happen at a lower layer.

    The second part of Fixture object creation is crafting the data.  Tying data
    to a Fixture class should be done with ``torment.fixtures.register``.  It
    provides a declarative interface that binds a dictionary to a Fixture (keys
    of dictionary become Fixture properties).  ``torment.fixtures.register``
    creates a subclass that the rest of the torment knows how to transform into
    test cases that are compatible with nose.

    Examples
    --------

    TODO ADD EXAMPLES

    Properties
    ----------

    * ``category``
    * ``description`` (override)
    * ``name`` (do **not** override)

    Methods To Override
    -------------------

    * ``__init__``
    * ``check``
    * ``initialize``
    * ``run (required)``
    * ``setup``

    Instance Variables
    ------------------

    :``context``: the ``torment.TestContext`` this case is running in which
                  provides the assertion methods of ``unittest.TestCase``.

    '''

    def __init__(self, context: 'torment.TestContext') -> None:
        '''Create Fixture

        Initializes the Fixture's context (can be changed like any other
        property).

        Parameters
        ----------

        :``context``: a subclass of ``torment.TestContext`` that provides
                      assertion methods and any other environmental information
                      for this test case

        '''

        self.context = context

    @property
    def category(self) -> str:
        '''Fixture's category (the containing testing module name)

        Examples
        --------

        :module:   test_torment.test_unit.test_fixtures.fixture_a44bc6dda6654b1395a8c2cbd55d964d
        :category: fixtures

        '''

        return self.__module__.rsplit('.', 2)[-2].replace('test_', '')

    @property
    def description(self) -> str:
        '''Test name in nose output (intended to be overridden).'''

        return '{0.uuid.hex}—{1}'.format(self, self.context.module)

    @property
    def name(self) -> str:
        '''Method name in nose runtime.'''

        return 'test_' + self.__class__.__name__

    def initialize(self) -> None:
        '''Post-data population initialization hook.

        .. note::
            Override as necessary.  Default provided so re-defenition is not
            necessary.

        Called during ``__init__`` and after properties have been populated by
        ``torment.fixtures.register``.


        '''

        pass

    def setup(self) -> None:
        '''Pre-run initialization hook.

        .. note::
            Override as necessary.  Default provided so re-defenition is not
            necessary.

        Called after properties have been populated by
        ``torment.fixtures.register``.

        '''

        pass

    def check(self) -> None:
        '''Check that run ran as expected.

        .. note::
            Override as necessary.  Default provided so re-defenition is not
            necessary.

        Called after ``run`` and should be used to verify that run performed the
        expected actions.

        '''

        pass

    def _execute(self) -> None:
        '''Run Fixture actions (setup, run, check).

        Core test loop for Fixture.  Executes setup, run, and check in order.

        '''

        self.setup()
        self.run()
        self.check()


@decorators.log
def of(fixture_classes: Iterable[type], context: Union[None, 'torment.TestContext'] = None) -> Iterable['torment.fixtures.Fixture']:
    '''Obtain all Fixture objects of the provided classes.

    Parameters
    ----------

    :``fixture_classes``: classes inheriting from ``torment.fixtures.Fixture``
    :``context``:         a ``torment.TestContext`` to initialize Fixtures with

    Return Value(s)
    ---------------

    Instantiated ``torment.fixtures.Fixture`` objects for each individual
    fixture class that inherits from one of the provided classes.

    '''

    classes = list(copy.copy(fixture_classes))
    fixtures = []  # type: Iterable[torment.fixtures.Fixture]

    while len(classes):
        current = classes.pop()
        subclasses = current.__subclasses__()

        if len(subclasses):
            classes.extend(subclasses)
        elif current not in fixture_classes:
            fixtures.append(current(context))

    return fixtures


def register(namespace, base_classes: Tuple[type], properties: Dict[str,Any]) -> None:
    '''Register a Fixture class in namespace with the given properties.

    Creates a Fixture class (not object) and inserts it into the provided
    namespace.  The properties is a dict but allows functions to reference other
    properties and acts like a small DSL (domain specific language).  This is
    really just a declarative way to compose data about a test fixture and make
    it repeatable.

    Files calling this function are expected to house one or more Fixtures and
    have a name that ends with a UUID without its hyphens.  For example:
    foo_38de9ceec5694c96ace90c9ca37e5bcb.py.  This UUID is used to uniquely
    track the Fixture through the test suite and allow Fixtures to scale without
    concern.

    Parameters
    ----------

    :``namespace``:    dictionary to insert the generated class into
    :``base_classes``: list of classes the new class should inherit
    :``properties``:   dictionary of properties with their values

    Properties can have the following forms:

    :functions: invoked with the Fixture as it's argument
    :classes:   instantiated without any arguments (unless it subclasses
                ``torment.fixtures.Fixture`` in which case it's passed context)
    :literals:  any standard python type (i.e. int, str, dict)

    .. note::
        function execution may error (this will be emitted as a logging event).
        functions will continually be tried until they resolve or the same set
        of functions is continually erroring.  These functions that failed to
        resolve are left in tact for later processing.

    Properties by the following names also have defined behavior:

    :description: added to the Fixture's description as an addendum
    :error:       must be a dictionary with three keys:
                  :class:  class to instantiate (usually an exception)
                  :args:   arguments to pass to class initialization
                  :kwargs: keyword arguments to pass to class initialization

    Properties by the following names are reserved and should not be used:

    * name

    '''

    props = copy.deepcopy(properties)  # ensure we have a clean copy of the data
                                       # and won't stomp on re-uses elsewhere in
                                       # someone's code

    desc = props.pop('description', None)  # type: Union[str, None]

    caller_frame = inspect.stack()[1]

    caller_file = caller_frame[1]
    caller_module = inspect.getmodule(caller_frame[0])

    my_uuid = uuid.UUID(os.path.basename(caller_file).replace('.py', '').rsplit('_', 1)[-1])
    class_name = _unique_class_name(namespace, my_uuid)

    @property
    def description(self) -> str:
        _ = super(self.__class__, self).description

        if desc is not None:
            _ += '—' + desc

        return _

    def __init__(self, context: 'torment.TestContext') -> None:
        super(self.__class__, self).__init__(context)

        functions = {}

        for name, value in props.items():
            if name == 'error':
                self.error = value['class'](*value.get('args', ()), **value.get('kwargs', {}))
                continue

            if inspect.isclass(value):
                if issubclass(value, Fixture):
                    value = value(self.context)
                else:
                    value = value()

            if inspect.isfunction(value):
                functions[name] = value
                continue

            setattr(self, name, value)

        _resolve_functions(functions, self)

        self.initialize()

    namespace[class_name] = type(class_name, base_classes, {
        'description': description,
        '__init__': __init__,
        '__module__': caller_module,
        'uuid': my_uuid,
    })


def _resolve_functions(functions: Dict[str, Callable[[Any], Any]], fixture: Fixture) -> None:
        '''Apply functions and collect values as properties on fixture.

        Call functions and apply their values as properteis on fixture.
        Functions will continue to get applied until no more functions resolve.
        All unresolved functions are logged and the last exception to have
        occurred is also logged.  This function does not return but adds the
        results to fixture directly.

        Parameters
        ----------

        :``functions``: dict mapping function names (property names) to
                        callable functions
        :``fixture``:   Fixture to add values to

        '''

        exc_info = last_function = None
        function_count = float('inf')

        while function_count > len(functions):
            function_count = len(functions)

            for name, function in copy.copy(functions).items():
                logger.debug('name: %s', name)

                try:
                    setattr(fixture, name, copy.deepcopy(function(fixture)))
                    del functions[name]
                except:
                    exc_info = sys.exc_info()
                    logger.debug('exc_info: %s', exc_info)

                    last_function = name

        if len(functions):
            logger.warning('unprocessed Fixture properties: %s', ','.join(functions.keys()))
            logger.warning('last exception from %s.%s:', fixture.name, last_function, exc_info = exc_info)

            for name, function in copy.copy(functions).items():
                setattr(fixture, name, function)


def _unique_class_name(namespace: Dict[str, Any], uuid: uuid.UUID) -> str:
    '''Generate unique to namespace name for a class using uuid.

    Parameters
    ----------

    :``namespace``: the namespace to verify uniqueness against
    :``uuid``:      the "unique" portion of the name

    Return Value(s)
    ---------------

    A unique string (in namespace) using uuid.

    '''

    count = 0

    name = original_name = 'f_' + uuid.hex
    while name in namespace:
        count += 1
        name = original_name + '_' + str(count)

    return name



# Copyright 2015 Alex Brandt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import re
import logging
import typing  # flake8: noqa (use mypy typing)
import warnings

from typing import Any
from typing import Callable

from torment import decorators
from torment import fixtures

logger = logging.getLogger(__name__)


@property
def _module(self) -> str:
    '''Actual module name corresponding to this context's testing module.'''

    return re.sub(r'\.[^.]+', '', self.__module__.replace('test_', ''), 1)


class MetaContext(type):
    '''``torment.TestContext`` class creator.

    Generates all testing methods that correspond with the fixtures associated
    with a ``torment.TestContext``.  Also updates the definitions of
    ``mocks_mask`` and ``mocks`` to include the union of all involved classes
    in the creation process (all parent classes and the class being created).

    When creating a ``torment.TestContext`` subclass, ensure you specify this
    class as its metaclass to automatically generate test cases based on its
    ``fixture_classes`` property.

    '''

    module = _module

    def __init__(cls, name, bases, dct) -> None:
        super(MetaContext, cls).__init__(name, bases, dct)

        cls.mocks_mask = set().union(getattr(cls, 'mocks_mask', set()), *[ getattr(base, 'mocks_mask', set()) for base in bases ])
        cls.mocks = set().union(getattr(cls, 'mocks', set()), *[ getattr(base, 'mocks', set()) for base in bases ])

        def generate_case(fixture: fixtures.Fixture) -> Callable[[Any], None]:
            '''Generate a ``unittest.TestCase`` compatible test method.

            Parameters
            ----------

            :``fixture``: the fixture to transform into a ``unittest.TestCase``
                          compatible test method

            Return Value(s)
            ---------------

            An acceptable method that nose will execute as a test case.

            '''

            def case(self) -> None:
                fixture.context = self
                fixture._execute()

            case.__name__ = fixture.name
            case.__doc__ = fixture.description

            if len(cls.mocks_mask):
                case.__doc__ += '—unmocked:' + ','.join(sorted(cls.mocks_mask))

            return case

        if not hasattr(cls, 'fixture_classes'):
            warnings.warn('type object \'{0}\' has no attribute \'fixture_classes\'')
        else:
            for fixture in fixtures.of(cls.fixture_classes, context = cls):
                _ = generate_case(fixture)
                setattr(cls, _.__name__, _)


class TestContext(unittest.TestCase):
    '''Environment for Fixture execution.

    Provides convenience methods indicating the environment a Fixture is
    executing in.  This includes a references to the real module corresponding
    to the context's testing module as well as a housing for the assertion
    methods.

    Inherits most of its functionality from ``unittest.TestCase`` with a couple
    of additions.  TestContext does extend setUp.

    When used in conjunction with ``torment.MetaContext``, the
    ``fixture_classes`` property must be an iterable of subclasses of
    ``torment.fixtures.Fixture``.

    Properties
    ----------

    * ``module``

    Public Methods
    --------------

    * ``patch``

    Class Variables
    ---------------

    :``mocks_mask``: set of mocks to mask from being mocked
    :``mocks``:      set of mocks this TestContext provides

    '''

    mocks_mask = set()  # type: Set[str]
    mocks = set()  # type: Set[str]

    module = _module

    def setUp(self) -> None:
        super().setUp()

        logger.debug('self.mocks_mask: %s', self.mocks_mask)
        logger.debug('self.mocks: %s', self.mocks)

    @decorators.log
    def patch(self, name: str) -> None:
        '''Patch name with mock in actual module.

        Sets up mock objects for the given symbol in the actual module
        corresponding to this context's testing module.

        Parameters
        ----------

        :``name``: the symbol to mock—must exist in the actual module under test

        '''

        logger.debug('self.module: %s', self.module)

        _ = unittest.mock.patch(self.module + '.' + name)
        setattr(self, 'mocked_' + name.replace('.', '_').strip('_'), _.start())
        self.addCleanup(_.stop)



# Copyright 2015 Alex Brandt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import inspect
import logging
import typing  # flake8: noqa (use mypy typing)
import unittest
import uuid

from torment import fixtures

logger = logging.getLogger(__name__)


class FixturesCreateUnitTest(unittest.TestCase):
    def test_fixture_create_without_context(self) -> None:
        '''torment.fixtures.Fixture() → TypeError'''

        self.assertRaises(TypeError, fixtures.Fixture)

    def test_fixture_create_with_context(self) -> None:
        '''torment.fixtures.Fixture(context).context == context'''

        c = unittest.TestCase()
        f = fixtures.Fixture(c)

        self.assertEqual(f.context, c)


class FixturesPropertyUnitTest(unittest.TestCase):
    def setUp(self) -> None:
        self.c = unittest.TestCase()
        self.f = fixtures.Fixture(self.c)

    def test_fixture_category(self) -> None:
        '''torment.fixtures.Fixture(context).category == 'fixtures' '''

        self.f.__module__ = 'test_torment.test_unit.test_fixtures.fixture_a44bc6dda6654b1395a8c2cbd55d964d'

        self.assertEqual(self.f.category, 'fixtures')

    def test_fixture_description(self) -> None:
        '''torment.fixtures.Fixture(context).description == '94d7c58f6ee44683936c21cb84d1e458—torment.fixtures' '''

        self.f.context.module = 'fixtures'
        self.f.uuid = uuid.UUID('94d7c58f6ee44683936c21cb84d1e458')

        self.assertEqual(self.f.description, '94d7c58f6ee44683936c21cb84d1e458—fixtures')

    def test_fixture_name(self) -> None:
        '''torment.fixtures.Fixture(context).name == 'test_94d7c58f6ee44683936c21cb84d1e458' '''

        self.f.__class__.__name__ = '94d7c58f6ee44683936c21cb84d1e458'

        self.assertEqual(self.f.name, 'test_94d7c58f6ee44683936c21cb84d1e458')


class OfUnitTest(unittest.TestCase):
    def test_of_zero(self) -> None:
        '''torment.fixtures.of(()) == []'''

        self.assertEqual(len(fixtures.of(())), 0)

    def test_of_many_without_subclasses(self) -> None:
        '''torment.fixtures.of(( FixtureA, )) == []'''

        class FixtureA(object):
            def __init__(self, context) -> None:
                pass

        self.assertEqual(len(fixtures.of(( FixtureA, ))), 0)

    def test_of_many_with_subclasses(self) -> None:
        '''torment.fixtures.of(( FixtureA, )) == [ fixture_a, ]'''

        class FixtureA(object):
            def __init__(self, context) -> None:
                pass

        class FixtureB(FixtureA):
            pass

        result = fixtures.of(( FixtureA, ))

        self.assertEqual(len(result), 1)
        self.assertIsInstance(result[0], FixtureB)


class RegisterUnitTest(unittest.TestCase):
    def setUp(self) -> None:
        _ = unittest.mock.patch('torment.fixtures.inspect')
        mocked_inspect = _.start()
        self.addCleanup(_.stop)

        mocked_inspect.configure_mock(**{ 'isclass': inspect.isclass, 'isfunction': inspect.isfunction, })

        mocked_inspect.stack.return_value = ( None, ( None, 'test_unit/test_d43830e2e9624dd19c438b15250c5818.py', ), )
        
        class ContextStub(object):
            pass

        self.context = ContextStub()
        self.context.module = mocked_inspect.getmodule.return_value = 'stack'

        self.ns = {}  # type: Dict[str, Any]
        self.class_name = 'f_d43830e2e9624dd19c438b15250c5818'

    def test_zero_properties(self) -> None:
        '''torment.fixtures.register({}, (), {})'''

        fixtures.register(self.ns, ( fixtures.Fixture, ), {})

        _ = self.ns[self.class_name](self.context)

        self.assertEqual(_.uuid, uuid.UUID('d43830e2e9624dd19c438b15250c5818'))

    def test_one_literal_properties(self) -> None:
        '''torment.fixtures.register({}, (), { 'a': 'a', })'''

        fixtures.register(self.ns, ( fixtures.Fixture, ), { 'a': 'a', })

        _ = self.ns[self.class_name](self.context)

        self.assertEqual(_.a, 'a')

    def test_one_class_properties(self) -> None:
        '''torment.fixtures.register({}, (), { 'a': class, })'''

        class A(object):
            pass

        fixtures.register(self.ns, ( fixtures.Fixture, ), { 'a': A, })

        _ = self.ns[self.class_name](self.context)

        self.assertIsInstance(_.a, A)

    def test_one_fixture_class_properties(self) -> None:
        '''torment.fixtures.register({}, (), { 'a': fixture_class, })'''

        class A(fixtures.Fixture):
            pass

        fixtures.register(self.ns, ( fixtures.Fixture, ), { 'a': A, })

        _ = self.ns[self.class_name](self.context)

        self.assertIsInstance(_.a, A)
        self.assertEqual(_.a.context, self.context)

    def test_one_function_properties(self) -> None:
        '''torment.fixtures.register({}, (), { 'a': self → None, })'''

        def a(self) -> None: pass  # flake8: noqa (one line function in tests)

        fixtures.register(self.ns, ( fixtures.Fixture, ), { 'a': a, })

        _ = self.ns[self.class_name](self.context)

        self.assertIsNone(_.a)

    def test_description_property(self) -> None:
        '''torment.fixtures.register({}, (), { 'description': 'needle', })'''

        fixtures.register(self.ns, ( fixtures.Fixture, ), { 'description': 'needle', })

        _ = self.ns[self.class_name](self.context)

        self.assertEqual(_.description, 'd43830e2e9624dd19c438b15250c5818—stack—needle')

    def test_error_property(self) -> None:
        '''torment.fixtures.register({}, (), { 'error': …, })'''

        fixtures.register(self.ns, ( fixtures.Fixture, ), { 'error': { 'class': RuntimeError, }, })

        _ = self.ns[self.class_name](self.context)

        self.assertIsInstance(_.error, RuntimeError)


class ResolveFunctionsUnitTest(unittest.TestCase):
    def setUp(self) -> None:
        class StubFixture(object):
            pass

        self.f = StubFixture()
        self.f.name = 'testing_fixture_stub'

        self.o = copy.deepcopy(self.f)

    def test_zero_functions(self) -> None:
        '''torment.fixtures._resolve_functions({}, fixture)'''

        fixtures._resolve_functions({}, self.f)

        self.assertEqual(dir(self.o), dir(self.f))

    def test_one_functions_without_parameters(self) -> None:
        '''torment.fixtures._resolve_functions({ 'a': ø → None, }, fixture)'''

        def a() -> None: pass  # flake8: noqa (one line function in tests)

        fixtures._resolve_functions({ 'a': a, }, self.f)

        self.assertEqual(id(self.f.a), id(a))

    def test_one_functions_with_self_parameter(self) -> None:
        '''torment.fixtures._resolve_functions({ 'a': self → None, }, fixture)'''

        def a(self) -> None: pass  # flake8: noqa (one line function in tests)

        fixtures._resolve_functions({ 'a': a, }, self.f)

        self.assertIsNone(self.f.a)

    def test_one_functions_with_self_parameter_raises_attributeerror(self) -> None:
        '''torment.fixtures._resolve_functions({ 'a': self → self.b, }, fixture)'''

        def a(self): return self.b  # flake8: noqa (one line function in tests)

        fixtures._resolve_functions({ 'a': a, }, self.f)

        self.assertEqual(id(self.f.a), id(a))

    def test_many_functions(self) -> None:
        '''torment.fixtures._resolve_functions({ 'a': self → self.b, 'b': self → None, }, fixture)'''

        def a(self) -> None: return self.b  # flake8: noqa (one line function in tests)
        def b(self) -> None: pass  # flake8: noqa (one line function in tests)

        fixtures._resolve_functions({ 'a': a, 'b': b, }, self.f)

        self.assertIsNone(self.f.a)
        self.assertIsNone(self.f.b)


class UniqueClassNameUnitTest(unittest.TestCase):
    def setUp(self) -> None:
        self.uuid = uuid.uuid4()

    def test_empty_namespace(self) -> None:
        '''torment.fixtures._unique_class_name({}, uuid) == 'f_{uuid}' '''

        n = fixtures._unique_class_name({}, self.uuid)

        self.assertEqual(n, 'f_' + self.uuid.hex)

    def test_one_namespace(self) -> None:
        '''torment.fixtures._unique_class_name({ 'f_{uuid}': None, }, uuid) == 'f_{uuid}_1' '''

        n = fixtures._unique_class_name({ 'f_' + self.uuid.hex: None, }, self.uuid)

        self.assertEqual(n, 'f_' + self.uuid.hex + '_1')

    def test_two_namespace(self) -> None:
        '''torment.fixtures._unique_class_name({ 'f_{uuid}': None, 'f_{uuid}_1': None, }, uuid) == 'f_{uuid}_2' '''

        n = fixtures._unique_class_name({ 'f_' + self.uuid.hex: None, 'f_' + self.uuid.hex + '_1': None, }, self.uuid)

        self.assertEqual(n, 'f_' + self.uuid.hex + '_2')



# Copyright 2015 Alex Brandt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

AUTHOR = 'Alex Brandt'
AUTHOR_EMAIL = 'alunduil@alunduil.com'
COPYRIGHT = '2015'
DESCRIPTION = 'A Study in Fixture Based Testing Frameworking'
LICENSE = 'Apache-2.0'
NAME = 'torment'
URL = 'https://github.com/kumoru/torment'
VERSION = '1.0.0'



# Copyright 2015 Alex Brandt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
import os
import typing  # flake8: noqa (use mypy typing)
import unittest

from torment import fixtures
from torment import MetaContext
from torment import TestContext

from torment import helpers


class ExtendFixture(fixtures.Fixture):
    @property
    def description(self) -> str:
        return '{0.uuid.hex}—{1}.extend({{ {0.parameters[base]} }}, {{ {0.parameters[extension]} }}) == {{ {0.expected} }}'.format(self, self.context.module)

    def run(self) -> None:
        self.result = helpers.extend(self.parameters['base'], self.parameters['extension'])

    def check(self) -> None:
        self.context.assertEqual(self.expected, self.result)

helpers.import_directory(__name__, os.path.dirname(__file__))


class HelperUnitTest(TestContext, metaclass = MetaContext):
    fixture_classes = (
        ExtendFixture,
    )


class ImportDirectoryUnitTest(unittest.TestCase):
    def setUp(self) -> None:
        _ = unittest.mock.patch('helpers.importlib', wraps = importlib)
        self.mocked_importlib = _.start()
        self.addCleanup(_.stop)

    @unittest.skip('implementation needed')
    def test_import_zero_modules(self) -> None:
        '''torment.helpers.import_directory(module_base, './mock_import_directories/zero')'''

        self.fail('implement me')

        helpers.import_directory(self.__module__, 'mock_import_directories/zero')

        self.assertFalse(self.mocked_importlib.import_module.called)


class FilenamesToModulenamesUnitTest(unittest.TestCase):
    def test_zero_filenames(self) -> None:
        '''torment.helpers._filenames_to_modulenames([], 'test_helpers', 'test_helpers') == []'''

        mns = helpers._filenames_to_modulenames([], 'test_helpers', 'test_helpers')

        self.assertCountEqual([], mns)

    def test_one_filenames_proper(self) -> None:
        '''torment.helpers._filenames_to_modulenames([ 'test_helpers/extend_dc1fe29410fc4587bae0962d519f94ce.py', ], 'test_helpers', 'test_helpers') == [ test_helpers.extend_dc1fe29410fc4587bae0962d519f94ce', ]'''

        mns = helpers._filenames_to_modulenames([ 'test_helpers/extend_dc1fe29410fc4587bae0962d519f94ce.py', ], 'test_helpers', 'test_helpers')

        self.assertCountEqual([ 'test_helpers.extend_dc1fe29410fc4587bae0962d519f94ce', ], mns)

    def test_one_filenames_init(self) -> None:
        '''torment.helpers._filenames_to_modulenames([ 'test_helpers/__init__.py', ], 'test_helpers', 'test_helpers') == []'''

        mns = helpers._filenames_to_modulenames([], 'test_helpers', 'test_helpers')

        self.assertCountEqual([], mns)

    def test_one_filenames_without_py(self) -> None:
        '''torment.helpers._filenames_to_modulenames([ 'test_helpers/data.txt', ], 'test_helpers', 'test_helpers') == []'''

        mns = helpers._filenames_to_modulenames([], 'test_helpers', 'test_helpers')

        self.assertCountEqual([], mns)

    @unittest.skip('test known_symbols logic')
    def test_one_filenames_with_duplicate_symbols(self) -> None:
        '''torment.helpers._filenames_to_modulenames([ …, ], 'test_helpers', 'test_helpers') == [ …, ]'''

        pass



# Copyright 2015 Alex Brandt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from codecs import open
from setuptools import find_packages
from setuptools import setup

with open(os.path.join(os.path.dirname(__file__), 'torment', 'information.py'), 'r', encoding = 'utf-8') as fh:
    exec(fh.read(), globals(), locals())

PARAMS = {}

PARAMS['name'] = NAME  # flake8: noqa—provided by exec
PARAMS['version'] = VERSION  # flake8: noqa—provided by exec
PARAMS['description'] = DESCRIPTION  # flake8: noqa—provided by exec

with open(os.path.join(os.path.dirname(__file__), 'README.rst'), 'r', encoding = 'utf-8') as fh:
    PARAMS['long_description'] = fh.read()

PARAMS['url'] = URL  # flake8: noqa—provided by exec
PARAMS['author'] = AUTHOR  # flake8: noqa—provided by exec
PARAMS['author_email'] = AUTHOR_EMAIL  # flake8: noqa—provided by exec
PARAMS['license'] = LICENSE  # flake8: noqa—provided by exec

PARAMS['classifiers'] = [
    'Development Status :: 2 - Pre-Alpha',
    # 'Development Status :: 3 - Alpha',
    # 'Development Status :: 4 - Beta',
    # 'Development Status :: 5 - Production/Stable',
    # 'Development Status :: 6 - Mature',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: Apache Software License',
    'Natural Language :: English',
    'Operating System :: OS Independent',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3 :: Only',
    'Topic :: Software Development :: Libraries',
    'Topic :: Software Development :: Libraries :: Python Modules',
    'Topic :: Software Development :: Testing',
]

PARAMS['keywords'] = [
    'declarative',
    'fixtures',
    'framework',
    'scalable',
    'testing',
]

PARAMS['packages'] = find_packages(exclude = ( 'test_*', ))

PARAMS['install_requires'] = [
    'mypy-lang',
]

# ..note::
#     Documentation Requires:
#     * sphinx_rtd_theme

PARAMS['extras_require'] = {}

PARAMS['test_suite'] = 'nose.collector'
PARAMS['tests_require'] = [
    'coverage',
    'nose',
]

PARAMS['data_files'] = [
    ( 'share/doc/{P[name]}-{P[version]}'.format(P = PARAMS), (
        'README.rst',
    )),
]

setup(**PARAMS)



# Copyright 2015 Alex Brandt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from torment import fixtures

from test_torment.test_unit.test_helpers import ExtendFixture

fixtures.register(globals(), ( ExtendFixture, ), {
    'parameters': {
        'base': {},
        'extension': {
            'a': 'a',
        },
    },

    'expected': {
        'a': 'a',
    },
})



# Copyright 2015 Alex Brandt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import inspect
import logging
import typing  # flake8: noqa (use mypy typing)

from typing import Any
from typing import Callable

logger = logging.getLogger(__name__)
logger.propogate = False
logger.addHandler(logging.NullHandler())


def log(prefix = ''):
    '''Add start and stop logging messages to the function.

    Parameters
    ----------

    :``prefix``: a prefix for the function name (optional)

    '''

    function = None

    if inspect.isfunction(prefix):
        prefix, function = '', prefix

    def _(function):
        @functools.wraps(function)
        def wrapper(*args, **kwargs):  # TODO use inspect.signature
            format_args = (
                prefix + function.__name__,
                ', '.join(list(map(str, args)) + [ ' = '.join(map(str, item)) for item in kwargs.items() ]),
            )

            logger.info('STARTING: %s(%s)', *format_args)

            try:
                return function(*args, **kwargs)
            except:
                logger.exception('EXCEPTION: %s(%s)', *format_args)
                raise
            finally:
                logger.info('STOPPING: %s(%s)', *format_args)

        return wrapper

    if function is not None:
        _ = _(function)

    return _


def mock(name: str) -> Callable[[Any], None]:
    '''Setup properties indicating status of name mock.

    This is designed to decorate ``torment.TestContext`` methods and is used to
    provide a consistent interface for determining if name is mocked once and
    only once.

    Parameters
    ----------

    :``name``: symbol in context's module to mock

    Return Value(s)
    ---------------

    True if name is mocked; otherwise, False.  Also, creates a property on the
    method's self, is_mocked_name, with this value.

    '''

    def _(func):
        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            logger.info('STARTING: mock ' + name)

            is_mocked = False

            sanitized_name = name.replace('.', '_').strip('_')

            if name in self.mocks_mask:
                logger.info('STOPPING: mock ' + name + '—MASKED')
            elif getattr(self, 'is_mocked_' + sanitized_name, False):
                is_mocked = True

                logger.info('STOPPING: mock ' + name + '—EXISTS')
            else:
                func(self, *args, **kwargs)

                is_mocked = True

                logger.info('STOPPING: mock ' + name)

            setattr(self, 'is_mocked_' + sanitized_name, is_mocked)

            return is_mocked

        return wrapper

    return _
